//****************************************************************************
//**
//**    SYS_MAIN.CPP
//**    System Main
//**
//**    Copyright (c) 1998 Chris Hargrove
//**    Notice: Usage of any code in this file is subject to the rules
//**            described in the LICENSE.TXT file included in this directory.
//**            Reading, compiling, or otherwise using this code constitutes
//**            automatic acceptance of the rules in said text file.
//**
//****************************************************************************
//============================================================================
//    IMPLEMENTATION HEADERS
//============================================================================
#define SYS_DEFS_INCLUDE_WIN32 // for SYSW_Argc, SYSW_Argv,
                               // and SYSW_MessageBox
#include "sys_defs.h"
#include "sys_main.h"

//============================================================================
//    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================
#define SAFEMALLOC_MARKER 0x2b3d1234
#define SAFEMALLOC_USEMARKER // enable malloc/free marker verify

//============================================================================
//    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================
//============================================================================
//    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE DATA
//============================================================================
static float sys_baseTime;
static float sys_lastTime;

static dword testReceiverHandle, quitReceiverHandle;

//============================================================================
//    INTERFACE DATA
//============================================================================
char sys_programPath[SYS_MAX_PATH];
float sys_curTime, sys_frameTime;

//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
//============================================================================
static void UpdateFrameTimes();
static void ResetFrameTimes();

//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTIONS
//============================================================================
static void UpdateFrameTimes()
{
	sys_curTime = M_TimeFloat() - sys_baseTime;
	sys_frameTime = sys_curTime - sys_lastTime;
	sys_lastTime = sys_curTime;
}

static void ResetFrameTimes()
{
	sys_baseTime = M_TimeFloat();
	sys_lastTime = sys_curTime = sys_frameTime = 0.0;
}


// Test receiver functions

static bbool InputQuitReceiver(in_event_t *event)
{
	// quit when Ctrl+X is pressed
	if ((event->eventType == INEV_PRESS)
	 && (event->key == 'x') && (event->flags & INEVF_CTRL))
	{
		SYS_Quit();
	}
	return(0); // no other input matters to this receiver
}

static bbool InputTestReceiver(in_event_t *event)
{
	static bbool exclusive = 0;
	char *keyname;

	keyname = IN_NameForKey(event->key);
	if (!keyname)
		keyname = "";
	switch(event->eventType)
	{
	case INEV_MOUSEMOVE:
		// Uncomment this if you want to see it print out mouse moves.
		// Be aware the response will be a bit chuggy since the debugging
		// window can get very flooded with these messages, and it's on a
		// separate thread so it's not the speediest thing in the world
		
		//SYS_Debugf("MOUSEMOVE: At:%d,%d  Change:%d,%d\n",
		//	event->mouseFreeX, event->mouseFreeY,
		//	event->mouseDeltaX, event->mouseDeltaY);
		break;
	case INEV_PRESS:
		if (event->key == INKEY_ESCAPE)
			SYS_Debugf("PRESS:   %s     Remember, press Ctrl+X if you want to quit\n", keyname);
		else
			SYS_Debugf("PRESS:   %s\n", keyname);
		break;
	case INEV_DRAG:
		// These flood the debug window like crazy, but uncomment this if you like
		//SYS_Debugf("DRAG:    %s\n", keyname);
		break;
	case INEV_RELEASE:
		SYS_Debugf("RELEASE: %s  (time held: %.1f secs)\n", keyname, event->pressTimeDelta);
		break;
	}
	return(1); // catch all input
}

//============================================================================
//    INTERFACE FUNCTIONS
//============================================================================

// The "Big Three" control functions

// initialization
void SYS_Init()
{
	IN_Init();
	SYSW_InitDebug();
	
	// Add two input receivers, one for the regular test and one dedicated
	// to catching Ctrl-X, to demonstrate the receiver stack.
	
	// The test receiver goes first...
	testReceiverHandle = IN_AddReceiver(InputTestReceiver);
	// And the quit receiver gets added on top of it...
	quitReceiverHandle = IN_AddReceiver(InputQuitReceiver);

	if (!M_CmdLineOption("noalert", 0)) // example of command line option use
	{
		SYS_Alertf("Hi!  I'm a little user input test.  Press some keys and "
				   "mouse buttons, and I'll tell you what was pressed.\n"
				   "You might need to move the black debug window a bit so "
				   "you can see what it's printing (it'll only print\n"
				   "when the main window is in focus).  Don't be surprised "
				   "if the debug window's reaction time is a bit slow too.\n\n"
				   "When you're done, just hit Ctrl+X and I'll go away.\n\n"
				   "BTW, you can run this program with the -noalert "
				   "option to skip this message.");
	}
	
	ResetFrameTimes(); // initialize frame times from this point
}

// main loop
void SYS_Frame()
{
	UpdateFrameTimes();
	
	// process user input
	IN_Frame();
}

// de-initialization
void SYS_Shutdown()
{
	// Remove the input receivers.  Not really necessary since IN_Shutdown
	// will take care of it, but it's a good practice to clean up one's mess
	IN_RemoveReceiver(testReceiverHandle);
	IN_RemoveReceiver(quitReceiverHandle);
	
	SYSW_ShutdownDebug();
	IN_Shutdown();
}

// alert notification
void SYS_Alertf(char* text, ...)
{
	static char textBuffer[2048];
	va_list args;
	va_start(args, text);
	vsprintf(textBuffer, text, args);
	va_end(args);
	SYSW_MessageBox(APPTITLE " Notification", MB_OK, textBuffer);
}

// debugging window output
void SYS_Debugf(char* text, ...)
{
	static char textBuffer[1024];
	va_list args;
	va_start(args, text);
	vsprintf(textBuffer, text, args);
	va_end(args);
	SYSW_DebugString(textBuffer);
}

// exit without error
void SYS_Quit()
{
	SYS_Shutdown();
	exit(0);
}

// exit with fatal error
void SYS_Errorf(char *text, ...)
{
	static char textBuffer[1024];
	va_list args;
	va_start(args, text);
	vsprintf(textBuffer, text, args);
	va_end(args);
	SYSW_MessageBox("Fatal Error", MB_OK, textBuffer);
	SYS_Shutdown();
	exit(1);
}

// command line argument functions, call windows equivalents
byte SYS_Argc()
{
	return(SYSW_Argc());
}

char* SYS_Argv(byte index)
{
	return(SYSW_Argv(index));
}

// safe malloc and free wrappers, with optional marker verify (4 byte cost)
void* SYS_SafeMalloc(int size)
{
#ifdef SAFEMALLOC_USEMARKER	
	size += 4;
#endif
	void *res = malloc(size);
	if (!res)
		SYS_Errorf("SYS_SafeMalloc: Out of memory on %d bytes", size);
#ifdef SAFEMALLOC_USEMARKER	
	*((long *)res) = SAFEMALLOC_MARKER;
	return((long *)res+1);
#else
	return(res);
#endif
}

void SYS_SafeFree(void **ptr)
{
	long *p;
	if (!(*ptr))
		SYS_Errorf("SYS_SafeFree: Attempt to free null ptr");
	p = (long *)*ptr;
#ifdef SAFEMALLOC_USEMARKER	
	p -= 1;
	if (*p != SAFEMALLOC_MARKER)
		SYS_Errorf("SYS_SafeFree: Attempt to free non-SafeMalloc'ed pointer");
#endif
	free(p);
	*ptr = NULL;
}

//============================================================================
//    INTERFACE CLASS BODIES
//============================================================================

//****************************************************************************
//**
//**    END MODULE SYS_MAIN.CPP
//**
//****************************************************************************

